package net.lensfrex.rd.service.activity;

import cn.wustlinghang.mywust.util.StringUtil;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.github.benmanes.caffeine.cache.Cache;
import com.github.benmanes.caffeine.cache.Caffeine;
import com.github.yitter.idgen.YitIdHelper;
import net.lensfrex.rd.dao.mapper.activity.ActivityMapper;
import net.lensfrex.rd.data.entity.table.Activity;
import net.lensfrex.rd.data.model.request.admin.ActivityInfoDto;
import net.lensfrex.rd.data.model.response.PaginationWrapper;
import net.lensfrex.rd.exception.data.DataConflictException;
import net.lensfrex.rd.exception.data.DataFormatWrongException;
import net.lensfrex.rd.exception.data.DataNotExistsException;
import net.lensfrex.rd.service.course.CourseService;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.TimeUnit;

@Service
public class ActivityService {
    private static final Cache<String, Activity> cache = Caffeine.newBuilder()
            .expireAfterAccess(1, TimeUnit.DAYS)
            .initialCapacity(16)
            .maximumSize(64)
            .build();

    private static final Cache<String, Set<Activity>> termCache = Caffeine.newBuilder()
            .expireAfterAccess(1, TimeUnit.DAYS)
            .initialCapacity(16)
            .maximumSize(64)
            .build();

    private final ActivityMapper activityMapper;

    private final CourseService courseService;

    public ActivityService(ActivityMapper activityMapper,
                           CourseService courseService) {
        this.activityMapper = activityMapper;
        this.courseService = courseService;
    }

    public PaginationWrapper<List<Activity>> list(int page, int pageSize) {
        Page<Activity> queryPage = new Page<>(page, pageSize);
        var result = activityMapper
                .selectPage(queryPage, new QueryWrapper<Activity>()
                        .select("id, name, term, start_time, end_time, create_time, edit_time")
                        .orderByDesc("end_time")
                        .eq("deleted", false)
                );

        return new PaginationWrapper<>(
                (int) result.getCurrent(),
                (int) result.getSize(),
                (int) result.getTotal(),
                result.getRecords()
        );
    }

    public Activity getById(String id) {
        return getById(id, true);
    }

    public Activity getById(String id, boolean useCache) {
        var activity = cache.getIfPresent(id);
        if (!useCache || activity == null) {
            activity = activityMapper.selectOne(new QueryWrapper<Activity>()
                    .select("id, name, term, start_time, end_time, create_time, edit_time")
                    .eq("id", id)
                    .eq("deleted", false)
                    .last("limit 1")
            );

            putCache(activity);
        }

        return activity;
    }

    public Set<Activity> getByTerm(String term) {
        var list = termCache.getIfPresent(term);
        if (list == null) {
            var result = activityMapper.selectList(new QueryWrapper<Activity>()
                    .select("id, name, term, start_time, end_time, create_time, edit_time")
                    .orderByDesc("end_time")
                    .eq("term", term)
                    .eq("deleted", false)
            );
            list = new HashSet<>(result);
            termCache.put(term, list);
        }

        return list;
    }

    private void putCache(Activity activity) {
        cache.put(activity.getId(), activity);
        var list = termCache.getIfPresent(activity.getTerm());
        if (list == null) {
            list = new HashSet<>(4);
        }

        list.add(activity);
        termCache.put(activity.getTerm(), list);
    }

    public List<Activity> getCurrentActivity() {
        var now = LocalDateTime.now();//.format(dateTimeFormatter);
        return activityMapper.selectList(new QueryWrapper<Activity>()
                .select("id, name, term, start_time, end_time, create_time, edit_time")
                .le("start_time", now)
                .ge("end_time", now)
                .eq("term", StringUtil.getCurrentTermString())
                .orderByDesc("end_time")
                .eq("deleted", false)
        );
    }

    private boolean checkConflict(String id, ActivityInfoDto dto) {
        // 检查开始时间和结束时间
        if (!dto.start().isBefore(dto.end())) {
            throw new DataFormatWrongException();
        }

        // 检查冲突
        var activities = activityMapper.selectList(new QueryWrapper<Activity>()
                .eq("term", dto.term())
                .eq("deleted", false)
        );

        for (var activity : activities) {
            if (activity.getId().equals(id)) {
                continue;
            }

            boolean conflict =
                    // ____|____|____i____i____
                    !dto.start().isAfter(activity.getEndTime()) &&
                            // ____i____i____|____|____
                            !dto.end().isBefore(activity.getStartTime());
            if (conflict) {
                return true;
            }
        }

        return false;
    }

    public void addActivity(ActivityInfoDto request) {
        if (checkConflict(null, request)) {
            throw new DataConflictException();
        }

        String id = Long.toString(YitIdHelper.nextId());

        Activity activity = new Activity();
        activity.setId(id);
        activity.setName(request.name());
        activity.setTerm(request.term());
        activity.setStartTime(request.start());
        activity.setEndTime(request.end());

        activityMapper.insert(activity);

        putCache(activity);
        courseService.invalidStudentCandidateCourseCache();
    }

    @Transactional
    public void delete(String id) {
        Activity activity = getById(id);
        activity.setDeleted(true);
        activity.setEditTime(LocalDateTime.now());

        activityMapper.update(activity, new QueryWrapper<Activity>()
                .eq("id", id)
                .eq("deleted", false)
        );

        cache.invalidate(id);
        termCache.invalidate(activity.getTerm());
        courseService.invalidStudentCandidateCourseCache();
    }

    @Transactional
    public Activity update(String id, ActivityInfoDto request) {
        Activity activity = getById(id);
        if (activity == null) {
            throw new DataNotExistsException();
        }

        if (checkConflict(id, request)) {
            throw new DataConflictException();
        }

        activity.setName(request.name());
        activity.setTerm(request.term());
        activity.setStartTime(request.start());
        activity.setEndTime(request.end());

        activityMapper.update(activity, new QueryWrapper<Activity>()
                .eq("id", id)
                .eq("deleted", false)
        );

        courseService.invalidStudentCandidateCourseCache();

        return this.getById(id, false);
    }
}
