package net.lensfrex.rd.service.course;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.github.benmanes.caffeine.cache.Cache;
import com.github.benmanes.caffeine.cache.Caffeine;
import com.github.yitter.idgen.YitIdHelper;
import net.lensfrex.rd.dao.mapper.course.CourseMapper;
import net.lensfrex.rd.data.dto.course.CourseDTO;
import net.lensfrex.rd.data.entity.table.Course;
import net.lensfrex.rd.data.model.response.PaginationWrapper;
import net.lensfrex.rd.service.user.UserInfoService;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

@Service
public class CourseService {
    private static final Cache<String, Course> cache = Caffeine.newBuilder()
            .expireAfterAccess(1, TimeUnit.HOURS)
            .initialCapacity(16)
            .maximumSize(64)
            .build();

    private static final Cache<String, ArrayList<CourseDTO>> studentCandidateCourseCache = Caffeine.newBuilder()
            .expireAfterAccess(1, TimeUnit.HOURS)
            .initialCapacity(16)
            .maximumSize(64)
            .build();

    private final CourseMapper courseMapper;

    private final UserInfoService userInfoService;

    public CourseService(CourseMapper courseMapper,
                         UserInfoService userInfoService) {
        this.courseMapper = courseMapper;
        this.userInfoService = userInfoService;
    }

    /**
     * 添加一个课程信息
     *
     * @param courseRequest 添加请求
     * @return 已添加的课程信息
     */
    @Transactional
    public CourseDTO add(CourseDTO courseRequest) {
        String id = String.valueOf(YitIdHelper.nextId());

        Course course = new Course();
        course.setId(id);
        course.setName(courseRequest.getName());
        course.setCourseHours(courseRequest.getCourseHours());
        course.setSpringTerm(courseRequest.getSpringTerm());
        course.setTargetMajor(courseRequest.getTargetMajor());
        course.setTargetGrade(courseRequest.getTargetGrade());

        courseMapper.insert(course);

        studentCandidateCourseCache.invalidateAll();
        return this.getCourse(id, false);
    }

    /**
     * 删除指定id的课程
     *
     * @param id 欲删除的课程id
     */
    @Transactional
    public void delete(String id) {
        Course course = new Course();
        course.setDeleted(true);
        course.setEditTime(LocalDateTime.now());
        courseMapper.update(course, new QueryWrapper<Course>()
                .eq("id", id)
                .eq("deleted", false)
        );

        cache.invalidate(id);
    }

    public CourseDTO getCourse(String id) {
        return getCourse(id, true);
    }

    /**
     * 根据课程id获取课程信息
     *
     * @param id       课程id
     * @param useCache 是否使用缓存
     * @return 课程信息
     */
    public CourseDTO getCourse(String id, boolean useCache) {
        var result = cache.getIfPresent(id);
        if (!useCache || result == null) {
            result = courseMapper.selectOne(new QueryWrapper<Course>()
                    .select("id", "name", "course_hours", "spring_term", "target_major", "target_grade")
                    .eq("id", id)
                    .eq("deleted", false)
            );
            if (result == null) {
                return null;
            }

            cache.put(id, result);
        }

        return convert(result);
    }

    /**
     * 分页获取所有课程
     *
     * @param page     第几页
     * @param pageSize 每页数据量
     * @return 分页包装数据
     */
    public PaginationWrapper<List<CourseDTO>> list(int page, int pageSize) {
        Page<Course> queryPage = new Page<>(page, pageSize);
        var result = courseMapper.selectPage(queryPage, new QueryWrapper<Course>()
                .select("id", "name", "course_hours", "spring_term", "target_major", "target_grade")
                .eq("deleted", false)
        );

        List<CourseDTO> courseDTOList = new ArrayList<>();
        var records = result.getRecords();
        for (var course : records) {
            courseDTOList.add(convert(course));
        }

        return new PaginationWrapper<>(
                (int) result.getCurrent(),
                (int) result.getSize(),
                (int) result.getTotal(),
                courseDTOList
        );
    }

    private CourseDTO convert(Course course) {
        CourseDTO courseDTO = new CourseDTO();
        courseDTO.setId(course.getId());
        courseDTO.setName(course.getName());
        courseDTO.setCourseHours(course.getCourseHours());
        courseDTO.setTargetMajor(course.getTargetMajor());
        courseDTO.setTargetGrade(course.getTargetGrade());
        courseDTO.setSpringTerm(course.getSpringTerm());

        return courseDTO;
    }

    private Course convert(CourseDTO courseDTO) {
        Course course = new Course();
        course.setName(courseDTO.getName());
        course.setCourseHours(courseDTO.getCourseHours());
        course.setSpringTerm(courseDTO.getSpringTerm());
        course.setTargetMajor(courseDTO.getTargetMajor());
        course.setTargetGrade(courseDTO.getTargetGrade());

        return course;
    }

    /**
     * 更新指定id的课程信息
     *
     * @param id  课程id
     * @param dto 新的课程信息
     * @return 更新好的课程信息
     */
    @Transactional
    public CourseDTO update(String id, CourseDTO dto) {
        var course = convert(dto);
        course.setEditTime(LocalDateTime.now());

        courseMapper.update(course, new QueryWrapper<Course>()
                .eq("id", id)
                .eq("deleted", false)
        );

        studentCandidateCourseCache.invalidateAll();
        return this.getCourse(id, false);
    }

    private boolean isSpringTerm() {
        int month = LocalDateTime.now().getMonthValue();

        // 3~7月认为是春季期
        return month < 8 && month > 2;
    }

    private int getGrade(Integer sessionYear) {
        var now = LocalDateTime.now();
        int month = now.getMonthValue();
        int year = now.getYear();

        // 到了秋季期的时间之后已经是下一个年级了
        if (month >= 8 || month < 2) {
            return year - sessionYear + 1;
        } else {
            return year - sessionYear;
        }
    }

    /**
     * 获取学生当前待选的课程
     *
     * @param uid 学生uid
     * @return 待选的课程列表
     */
    public ArrayList<CourseDTO> getCandidateCourseForStudent(String uid) {
        var result = studentCandidateCourseCache.getIfPresent(uid);
        if (result != null) {
            return result;
        }

        var courseInfo = userInfoService.getAsStudent(uid);

        var grade = getGrade(courseInfo.getGrade());
        var courses = courseMapper.getCourseByMajorAndGrade(courseInfo.getMajor(), grade, isSpringTerm());

        result = new ArrayList<>(courses.size());
        for (var course : courses) {
            result.add(convert(course));
        }

        studentCandidateCourseCache.put(uid, result);
        return result;
    }

    public void invalidStudentCandidateCourseCache() {
        studentCandidateCourseCache.invalidateAll();
    }
}
