package net.lensfrex.rd.service.major;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.github.benmanes.caffeine.cache.Cache;
import com.github.benmanes.caffeine.cache.Caffeine;
import com.github.yitter.idgen.YitIdHelper;
import net.lensfrex.rd.dao.mapper.major.MajorMapper;
import net.lensfrex.rd.data.entity.table.Major;
import net.lensfrex.rd.exception.data.DataNotExistsException;
import org.springframework.stereotype.Service;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.TimeUnit;

@Service
public class MajorService {
    private static final Cache<String, Major> cache = Caffeine.newBuilder()
            .expireAfterAccess(30, TimeUnit.DAYS)
            .initialCapacity(16)
            .maximumSize(64)
            .build();

    private final MajorMapper majorMapper;

    public MajorService(MajorMapper majorMapper) {
        this.majorMapper = majorMapper;
    }

    public Major addMajorIfNotExists(String name) {
        Major major = this.getMajorByName(name);
        if (major == null) {
            major = new Major();
            String id = Long.toString(YitIdHelper.nextId());
            major.setId(id);
            major.setName(name);

            majorMapper.insert(major);

            major = getMajor(id, false);

            putCache(major);
        }

        return major;
    }

    public void deleteMajor(String id) {
        var major = majorMapper.selectOne(new QueryWrapper<Major>()
                .select("name")
                .eq("id", id)
                .eq("deleted", false)
                .last("limit 1")
        );

        if (major == null) {
            return;
        }

        major.setDeleted(true);
        major.setEditTime(LocalDateTime.now());

        majorMapper.update(major, new QueryWrapper<Major>()
                .eq("id", id)
                .eq("deleted", false)
        );

        cache.invalidate(id);
        cache.invalidate(major.getName());
    }

    public Major getMajor(String id, boolean useCache) {
        var major = cache.getIfPresent(id);
        if (useCache && major != null) {
            return major;
        }

        major = majorMapper.selectOne(new QueryWrapper<Major>()
                .eq("id", id)
                .eq("deleted", false)
        );

        if (major == null) {
            throw new DataNotExistsException();
        }

        putCache(major);

        return major;
    }

    public Major getMajor(String id) {
        return getMajor(id, true);
    }

    public Major getMajorByName(String name) {
        var major = cache.getIfPresent(name);
        if (major == null) {
            major = majorMapper.selectOne(new QueryWrapper<Major>()
                    .eq("name", name)
                    .eq("deleted", false)
                    .last("limit 1")
            );
        }

        return major;
    }

    public Set<Major> listAllMajor() {
        var cacheMap = cache.asMap();

        if (!cacheMap.isEmpty()) {
            return new HashSet<>(cacheMap.values());
        }

        var results = majorMapper.selectList(new QueryWrapper<Major>()
                .select("id", "name")
                .eq("deleted", false)
        );

        for (var result : results) {
            putCache(result);
        }

        return new HashSet<>(results);
    }

    private void putCache(Major major) {
        cache.put(major.getId(), major);
        cache.put(major.getName(), major);
    }
}
