package net.lensfrex.rd.service.schedule;
import com.alibaba.excel.EasyExcel;
import com.alibaba.excel.ExcelWriter;
import com.alibaba.excel.util.MapUtils;
import com.alibaba.excel.write.metadata.WriteSheet;
import com.alibaba.excel.write.metadata.fill.FillConfig;
import com.google.common.collect.Maps;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.github.benmanes.caffeine.cache.Cache;
import com.github.benmanes.caffeine.cache.Caffeine;
import com.github.yitter.idgen.YitIdHelper;
import lombok.extern.slf4j.Slf4j;
import net.lensfrex.rd.dao.mapper.schedule.CourseScheduleMapper;
import net.lensfrex.rd.dao.mapper.schedule.CourseScheduleSelectionMapper;
import net.lensfrex.rd.dao.mapper.schedule.TeacherClassMapper;
import net.lensfrex.rd.data.dto.schedule.CourseScheduleDTO;
import net.lensfrex.rd.data.dto.schedule.StudentExportDto;
import net.lensfrex.rd.data.entity.table.*;
import net.lensfrex.rd.data.model.response.PaginationWrapper;
import net.lensfrex.rd.exception.data.DataNotExistsException;
import net.lensfrex.rd.service.activity.ActivityService;
import net.lensfrex.rd.service.user.StudentUserService;
import net.lensfrex.rd.service.user.TeacherUserService;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringJoiner;
import java.util.concurrent.TimeUnit;

@Slf4j
@Service
public class CourseScheduleService {
    private static final Cache<String, CourseSchedule> cache = Caffeine.newBuilder()
            .expireAfterAccess(1, TimeUnit.HOURS)
            .initialCapacity(16)
            .maximumSize(64)
            .build();

    private static final Cache<String, List<CourseSchedule>> courseScheduleCache = Caffeine.newBuilder()
            .expireAfterAccess(1, TimeUnit.HOURS)
            .initialCapacity(16)
            .maximumSize(64)
            .build();

    private final CourseScheduleMapper courseScheduleMapper;
    private final CourseScheduleSelectionMapper courseScheduleSelectionMapper;
    private final TeacherClassMapper teacherClassMapper;
    private final TeacherUserService teacherUserService;

    private final StudentUserService studentUserService;

    private final ActivityService activityService;

    public CourseScheduleService(CourseScheduleMapper courseScheduleMapper,
                                 CourseScheduleSelectionMapper courseScheduleSelectionMapper,
                                 TeacherClassMapper teacherClassMapper,
                                 TeacherUserService teacherUserService,
                                 StudentUserService studentUserService,
                                 ActivityService activityService) {

        this.courseScheduleMapper = courseScheduleMapper;
        this.courseScheduleSelectionMapper = courseScheduleSelectionMapper;
        this.teacherClassMapper = teacherClassMapper;
        this.teacherUserService = teacherUserService;
        this.studentUserService = studentUserService;
        this.activityService = activityService;
    }

    public CourseScheduleDTO getScheduleDTO(String id) {
        var schedule = getSchedule(id);

        var result = convert(schedule);
        var teachers = this.getScheduleTeacherInfo(id);

        result.setTeachers(teachers);

        return result;
    }

    public CourseSchedule getSchedule(String id) {
        return getSchedule(id, true);
    }

    public CourseSchedule getSchedule(String id, boolean useCache) {
        var schedule = cache.getIfPresent(id);
        if (!useCache || schedule == null) {
            schedule = courseScheduleMapper.selectOne(new QueryWrapper<CourseSchedule>()
                    .select("id", "name", "course", "start_week", "end_week", "time", "week_day",
                            "academic_term", "place", "ratio", "distribution")
                    .eq("id", id)
                    .eq("deleted", false)
                    .last("limit 1")
            );
            if (schedule == null) {
                throw new DataNotExistsException();
            }

            putCache(schedule);
        }

        return schedule;
    }

    private List<CourseSchedule> getScheduleOfCourse(String courseId, String term) {
        String key = new StringJoiner(":").add(courseId).add(term).toString();
        var result = courseScheduleCache.getIfPresent(key);
        if (result == null) {
            result = courseScheduleMapper.selectList(new QueryWrapper<CourseSchedule>()
                    .select("id", "name", "course", "start_week", "end_week", "time", "week_day",
                            "academic_term", "place", "ratio", "distribution")
                    .eq("academic_term", term)
                    .eq("course", courseId)
                    .eq("deleted", false)
            );

            courseScheduleCache.put(key, result);
        }

        return result;
    }

    public List<CourseScheduleDTO> getTermCourseSchedule(String courseId, String term) {
        var schedules = getScheduleOfCourse(courseId, term);

        var dto = new ArrayList<CourseScheduleDTO>(schedules.size());
        for (var schedule : schedules) {
            dto.add(convert(schedule));
        }

        return dto;
    }

    public PaginationWrapper<List<CourseScheduleDTO>> getScheduleOfCourse(String courseId, int page, int pageSize) {
        Page<CourseSchedule> queryPage = new Page<>(page, pageSize);
        var pageRecord = courseScheduleMapper.selectPage(queryPage, new QueryWrapper<CourseSchedule>()
                .select("id", "name", "course", "start_week", "end_week", "time", "week_day",
                        "academic_term", "place", "ratio", "distribution")
                .eq("course", courseId)
                .eq("deleted", false)
        );

        return wrapPaginationResult(pageRecord);
    }

    public List<CourseScheduleDTO> getStudentTermSchedule(String uid, String term) {
        var activityList = activityService.getByTerm(term);
        if (activityList.isEmpty()) {
            return new ArrayList<>();
        }

        var activityIds = new ArrayList<String>();
        for (var activity : activityList) {
            // 只取已经开始的活动
            if (LocalDateTime.now().isAfter(activity.getStartTime())) {
                activityIds.add(activity.getId());
            }
        }

        var records = courseScheduleSelectionMapper.selectList(new QueryWrapper<CourseSelection>()
                .eq("student", uid)
                .in("activity", activityIds)
                .eq("deleted", false)
        );

        if (records.isEmpty()) {
            return new ArrayList<>();
        }

        var scheduleIds = new ArrayList<String>(records.size());

        for (var record : records) {
            scheduleIds.add(record.getSchedule());
        }

        var schedules = courseScheduleMapper.selectList(new QueryWrapper<CourseSchedule>()
                .select("id", "name", "course", "start_week", "end_week", "time", "week_day",
                        "academic_term", "place", "ratio", "distribution")
                .in("id", scheduleIds)
                .eq("deleted", false)
        );

        var result = new ArrayList<CourseScheduleDTO>();
        for (var schedule : schedules) {
            result.add(convert(schedule));
        }

        return result;
    }

    private PaginationWrapper<List<CourseScheduleDTO>> wrapPaginationResult(Page<CourseSchedule> pageRecord) {
        var records = pageRecord.getRecords();
        var resultList = new ArrayList<CourseScheduleDTO>(records.size());
        for (var schedule : records) {
            resultList.add(convert(schedule));
        }

        return new PaginationWrapper<>(
                (int) pageRecord.getCurrent(),
                (int) pageRecord.getSize(),
                (int) pageRecord.getTotal(),
                resultList
        );
    }

    static CourseSchedule convert(CourseScheduleDTO dto) {
        CourseSchedule courseSchedule = new CourseSchedule();

        courseSchedule.setId(dto.getId());
        courseSchedule.setName(dto.getName());
        courseSchedule.setCourse(dto.getCourse());
        courseSchedule.setStartWeek(dto.getStartWeek());
        courseSchedule.setEndWeek(dto.getEndWeek());
        courseSchedule.setTime(dto.getTime());
        courseSchedule.setWeekDay(dto.getWeekDay());
        courseSchedule.setAcademicTerm(dto.getAcademicTerm());
        courseSchedule.setPlace(dto.getPlace());
        courseSchedule.setRatio(dto.getRatio());
        courseSchedule.setDistribution(dto.getDistribution());

        return courseSchedule;

    }

    static CourseScheduleDTO convert(CourseSchedule courseSchedule) {
        CourseScheduleDTO courseScheduleDTO = new CourseScheduleDTO();

        courseScheduleDTO.setId(courseSchedule.getId());
        courseScheduleDTO.setName(courseSchedule.getName());
        courseScheduleDTO.setCourse(courseSchedule.getCourse());
        courseScheduleDTO.setStartWeek(courseSchedule.getStartWeek());
        courseScheduleDTO.setEndWeek(courseSchedule.getEndWeek());
        courseScheduleDTO.setTime(courseSchedule.getTime());
        courseScheduleDTO.setWeekDay(courseSchedule.getWeekDay());
        courseScheduleDTO.setAcademicTerm(courseSchedule.getAcademicTerm());
        courseScheduleDTO.setPlace(courseSchedule.getPlace());
        courseScheduleDTO.setRatio(courseSchedule.getRatio());
        courseScheduleDTO.setDistribution(courseSchedule.getDistribution());

        return courseScheduleDTO;
    }

    private void putCache(CourseSchedule courseSchedule) {
        cache.put(courseSchedule.getId(), courseSchedule);

        String key = new StringJoiner(":")
                .add(courseSchedule.getCourse())
                .add(courseSchedule.getAcademicTerm())
                .toString();

        var list = courseScheduleCache.getIfPresent(key);
        if (list == null) {
            list = new ArrayList<>(8);
            list.add(courseSchedule);
            courseScheduleCache.put(key, list);
        }

        list.remove(courseSchedule);
        list.add(courseSchedule);
    }

    public List<String> getScheduleTeachers(String scheduleId) {
        var teacherClassList = teacherClassMapper.selectList(new QueryWrapper<TeacherClass>()
                .select("teacher")
                .eq("class", scheduleId)
                .eq("deleted", false)
        );

        if (teacherClassList == null) {
            return new ArrayList<>();
        }

        List<String> teachers = new ArrayList<>(teacherClassList.size());
        for (var result : teacherClassList) {
            teachers.add(result.getTeacher());
        }

        return teachers;
    }

    public List<TeacherInfo> getScheduleTeacherInfo(String scheduleId) {
        var teacherIds = this.getScheduleTeachers(scheduleId);

        return teacherUserService.getInfoById(teacherIds);
    }

    @Transactional
    public CourseSchedule add(CourseScheduleDTO dto) {
        String id = Long.toString(YitIdHelper.nextId());

        CourseSchedule courseSchedule = convert(dto);
        courseSchedule.setId(id);

        courseScheduleMapper.insert(courseSchedule);

        var teachers = dto.getTeacherIds();

        if (!teachers.isEmpty()) {
            teacherClassMapper.insetBatch(id, teachers);
        }

        return getSchedule(id, false);
    }

    @Transactional
    public CourseSchedule update(String id, CourseScheduleDTO dto) {
        CourseSchedule courseSchedule = convert(dto);
        courseSchedule.setId(id);

        courseScheduleMapper.update(courseSchedule, new QueryWrapper<CourseSchedule>()
                .eq("id", id)
                .eq("deleted", false)
        );

        var teachers = dto.getTeacherIds();

        var existsTeachers = teacherClassMapper.selectList(new QueryWrapper<TeacherClass>()
                        .eq("class", id)
                // deleted记录也一同带上，所以这里就不带deleted过滤了
        );

        var deletedTeacherIds = new ArrayList<String>(existsTeachers.size());
        for (var existsTeacher : existsTeachers) {
            if (!dto.getTeacherIds().contains(existsTeacher.getTeacher())) {
                deletedTeacherIds.add(existsTeacher.getTeacher());
            }
        }

        TeacherClass deleteEntity = new TeacherClass();
        deleteEntity.setEditTime(LocalDateTime.now());
        deleteEntity.setDeleted(true);

        if (!deletedTeacherIds.isEmpty()) {
            teacherClassMapper.update(deleteEntity, new QueryWrapper<TeacherClass>()
                    .in("teacher", deletedTeacherIds)
            );
        }

        if (!teachers.isEmpty()) {
            teacherClassMapper.insetBatch(id, teachers);
        }

        return getSchedule(id, false);
    }

    @Async
    @Transactional
    public void duplicate(String course, String sourceTerm, String targetTerm) {
        var classList = this.getScheduleOfCourse(course, sourceTerm);
        if (classList.isEmpty()) {
            return;
        }

        for (CourseSchedule clazz : classList) {
            var teachers = this.getScheduleTeachers(clazz.getId());

            String id = Long.toString(YitIdHelper.nextId());
            CourseSchedule clone = new CourseSchedule();
            clone.setId(id);
            clone.setName(clazz.getName());
            clone.setCourse(course);
            clone.setStartWeek(clazz.getStartWeek());
            clone.setEndWeek(clone.getEndWeek());
            clone.setTime(clone.getTime());
            clone.setWeekDay(clazz.getWeekDay());
            clone.setAcademicTerm(targetTerm);
            clone.setPlace(clazz.getPlace());
            clone.setRatio(clone.getRatio());
            clone.setDistribution(clazz.getDistribution());
            clone.setCreateTime(LocalDateTime.now());
            clone.setEditTime(LocalDateTime.now());
            clone.setDeleted(false);

            courseScheduleMapper.insert(clone);
            teacherClassMapper.insetBatch(id, teachers);
        }
    }

    public void delete(String id) {
        CourseSchedule schedule = getSchedule(id);
        schedule.setDeleted(true);
        schedule.setEditTime(LocalDateTime.now());

        courseScheduleMapper.update(schedule, new QueryWrapper<CourseSchedule>()
                .eq("id", id)
                .eq("deleted", false)
        );

        cache.invalidate(id);
        String key = new StringJoiner(":")
                .add(schedule.getCourse())
                .add(schedule.getAcademicTerm())
                .toString();
        courseScheduleCache.invalidate(key);
    }

    public int getSelectedCount(String scheduleId) {
        return courseScheduleSelectionMapper.selectCount(new QueryWrapper<CourseSelection>()
                .eq("schedule", scheduleId)
                .eq("deleted", false)
        ).intValue();
    }

    public byte[] exportStudentExcel(String scheduleId) {
        var courseSelectionResultList = courseScheduleSelectionMapper.selectList(new QueryWrapper<CourseSelection>()
                .select("student")
                .eq("schedule", scheduleId)
                .eq("deleted", false)
        );

        var studentIdList = new ArrayList<String>(courseSelectionResultList.size());
        for (var courseSelection : courseSelectionResultList) {
            studentIdList.add(courseSelection.getStudent());
        }

        var studentResultList = studentUserService.getByIds(studentIdList);

        List<StudentExportDto> studentExportDtoList = new ArrayList<>(studentResultList.size());
        for (int i = 0; i < studentResultList.size(); i++) {
            var student = studentResultList.get(i);
            studentExportDtoList.add(convert(i + 1, student));
        }

        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        InputStream template = getClass().getResourceAsStream("/template_student_list.xlsx");
        try (ExcelWriter excelWriter = EasyExcel.write(outputStream).withTemplate(template).build()) {
            WriteSheet writeSheet = EasyExcel.writerSheet().build();

            FillConfig fillConfig = FillConfig.builder().forceNewRow(Boolean.TRUE).build();
            excelWriter.fill(studentExportDtoList, fillConfig, writeSheet);

            Map<String, Object> map = MapUtils.newHashMap();
            var scheduleInfo = courseScheduleMapper.selectOne(new QueryWrapper<CourseSchedule>()
                    .eq("id", scheduleId)
                    .eq("deleted", false)
                    .last("limit 1")
            );

            map.put("courseName", scheduleInfo.getName());
            map.put("count", this.getSelectedCount(scheduleId));

            excelWriter.fill(map, writeSheet);

            excelWriter.finish();
        }

        return outputStream.toByteArray();
    }

    private StudentExportDto convert(int index, StudentInfo studentInfo) {
        StudentExportDto studentExportDto = new StudentExportDto();
        studentExportDto.setId(index);
        studentExportDto.setStudentNumber(studentInfo.getStudentNumber());
        studentExportDto.setName(studentInfo.getName());

        return studentExportDto;
    }
}
