package net.lensfrex.rd.service.score;

import com.alibaba.excel.EasyExcel;
import com.alibaba.excel.ExcelWriter;
import com.alibaba.excel.util.MapUtils;
import com.alibaba.excel.write.metadata.WriteSheet;
import com.alibaba.excel.write.metadata.fill.FillConfig;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.github.yitter.idgen.YitIdHelper;
import net.lensfrex.rd.dao.mapper.schedule.CourseScheduleMapper;
import net.lensfrex.rd.dao.mapper.schedule.CourseScheduleSelectionMapper;
import net.lensfrex.rd.dao.mapper.score.ScoreMapper;
import net.lensfrex.rd.data.dto.score.ScheduleScore;
import net.lensfrex.rd.data.dto.score.ScoreExportDto;
import net.lensfrex.rd.data.dto.score.StudentScore;
import net.lensfrex.rd.data.entity.table.CourseSchedule;
import net.lensfrex.rd.data.entity.table.CourseSelection;
import net.lensfrex.rd.data.entity.table.Score;
import net.lensfrex.rd.exception.data.DataFormatWrongException;
import net.lensfrex.rd.exception.data.DataNotExistsException;
import net.lensfrex.rd.service.user.StudentUserService;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

@Service
public class ScoreService {
    private final ScoreMapper scoreMapper;
    private final CourseScheduleSelectionMapper courseScheduleSelectionMapper;
    private final CourseScheduleMapper courseScheduleMapper;
    private final StudentUserService studentUserService;

    public ScoreService(ScoreMapper scoreMapper,
                        CourseScheduleSelectionMapper courseScheduleSelectionMapper,
                        CourseScheduleMapper courseScheduleMapper,
                        StudentUserService studentUserService) {

        this.scoreMapper = scoreMapper;
        this.courseScheduleSelectionMapper = courseScheduleSelectionMapper;

        this.courseScheduleMapper = courseScheduleMapper;
        this.studentUserService = studentUserService;
    }

    public List<StudentScore> getStudentScore(String uid) {
        return scoreMapper.getStudentScore(uid);
    }

    public List<ScheduleScore> getClassScore(String scheduleId) {
        return scoreMapper.getScheduleAllStudentScore(scheduleId);
    }

    @Transactional
    public void insertOrUpdateScore(String scheduleId, String scoreId, ScheduleScore scheduleScore) {
        if (scheduleScore.getUsualScore() == null || scheduleScore.getExamScore() == null || scheduleScore.getNumber() == null) {
            throw new DataFormatWrongException();
        }

        String studentUid = studentUserService.getByStudentNumber(scheduleScore.getNumber()).id();

        boolean isCurrentCourseScheduleStudent = courseScheduleSelectionMapper.exists(new QueryWrapper<CourseSelection>()
                .eq("student", studentUid)
                .eq("schedule", scheduleId)
                .eq("deleted", false)
        );
        if (!isCurrentCourseScheduleStudent) {
            throw new DataNotExistsException("不是当前课堂的学生");
        }

        if (scoreId == null) {
            scoreId = Long.toString(YitIdHelper.nextId());
        }

        Score score = convert(scheduleScore);
        score.setId(scoreId);
        score.setStudent(studentUid);
        score.setSchedule(scheduleId);
        score.setEditTime(LocalDateTime.now());

        scoreMapper.insertOrUpdateScore(score);
    }

    public byte[] exportScoreExcel(String scheduleId) {
        var scheduleInfo = courseScheduleMapper.selectOne(new QueryWrapper<CourseSchedule>()
                .select("name, ratio")
                .eq("id", scheduleId)
                .eq("deleted", false)
                .last("limit 1")
        );
        if (scheduleId == null) {
            return null;
        }

        var scoreList = scoreMapper.getScheduleAllStudentScore(scheduleId);

        double examRatio = scheduleInfo.getRatio() * 0.01;

        int count = 0;
        double sum = 0.0;

        List<ScoreExportDto> scoreExportDtoList = new ArrayList<>(scoreList.size());
        for (int i = 0; i < scoreList.size(); i++) {
            var score = scoreList.get(i);
            ScoreExportDto dto = convert(i + 1, score);
            if (score.getId() != null) {
                count++;
                double weightedUsualScore = score.getUsualScore() * (1 - examRatio);
                double weightedExamScore = score.getExamScore() * examRatio;
                double finalScore = weightedUsualScore + weightedExamScore;
                sum += finalScore;
                dto.setFinalScore(finalScore);
            }

            scoreExportDtoList.add(dto);
        }

        double avg = 0.0;
        if(count != 0) {
            avg = sum / count;
        }

        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        InputStream template = getClass().getResourceAsStream("/template_score.xlsx");
        try (ExcelWriter excelWriter = EasyExcel.write(outputStream).withTemplate(template).build()) {
            WriteSheet writeSheet = EasyExcel.writerSheet().build();

            FillConfig fillConfig = FillConfig.builder().forceNewRow(Boolean.TRUE).build();
            excelWriter.fill(scoreExportDtoList, fillConfig, writeSheet);

            Map<String, Object> map = MapUtils.newHashMap();

            map.put("courseName", scheduleInfo.getName());
            map.put("count", count);
            map.put("avg", avg);
            map.put("sum", sum);
            String ratioText = (1 - examRatio) + "/" + examRatio;
            map.put("ratio", ratioText);
            excelWriter.fill(map, writeSheet);

            excelWriter.finish();
        }

        return outputStream.toByteArray();
    }

    public static ScoreExportDto convert(int index, ScheduleScore scheduleScore) {
        ScoreExportDto scoreExportDto = new ScoreExportDto();
        scoreExportDto.setId(index);
        scoreExportDto.setName(scheduleScore.getName());
        scoreExportDto.setNumber(scheduleScore.getNumber());
        scoreExportDto.setUsualScore(scheduleScore.getUsualScore());
        scoreExportDto.setExamScore(scheduleScore.getExamScore());

        return scoreExportDto;
    }

    public static Score convert(ScheduleScore scheduleScore) {
        Score score = new Score();
        score.setId(scheduleScore.getId());
//        score.setStudent(score.getStudent());
        score.setUsualScore(scheduleScore.getUsualScore());
        score.setExamScore(scheduleScore.getExamScore());

        return score;
    }
}
