package net.lensfrex.rd.service.user;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.github.yitter.idgen.YitIdHelper;
import net.lensfrex.rd.dao.mapper.user.StudentInfoMapper;
import net.lensfrex.rd.data.dto.user.UserBasicInfoDTO;
import net.lensfrex.rd.data.dto.user.student.BasicStudentInfoDto;
import net.lensfrex.rd.data.dto.user.student.NewStudentDTO;
import net.lensfrex.rd.data.dto.user.student.StudentUserInfoDTO;
import net.lensfrex.rd.data.entity.table.StudentInfo;
import net.lensfrex.rd.data.enums.UserType;
import net.lensfrex.rd.data.model.response.PaginationWrapper;
import net.lensfrex.rd.exception.user.UserNotExistsException;
import net.lensfrex.rd.service.major.MajorService;
import net.lensfrex.rd.service.official.OfficialUndergradStudentInfoService;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;

@Service
public class StudentUserService {
    private final StudentInfoMapper studentInfoMapper;
    private final UserBasicService userBasicService;
    private final UserInfoService userInfoService;
    private final MajorService majorService;

    private final OfficialUndergradStudentInfoService officialUndergradStudentInfoService;

    public StudentUserService(StudentInfoMapper studentInfoMapper,
                              UserBasicService userBasicService,
                              UserInfoService userInfoService,
                              MajorService majorService,
                              OfficialUndergradStudentInfoService officialUndergradStudentInfoService) {

        this.studentInfoMapper = studentInfoMapper;
        this.userBasicService = userBasicService;
        this.userInfoService = userInfoService;
        this.majorService = majorService;

        this.officialUndergradStudentInfoService = officialUndergradStudentInfoService;
    }

    public PaginationWrapper<List<StudentInfo>> list(int page, int pageSize) {
        Page<StudentInfo> queryPage = new Page<>(page, pageSize);
        var result = studentInfoMapper.selectPage(queryPage, new QueryWrapper<StudentInfo>()
                .select("id, student_number, name, major, grade, class, create_time, edit_time")
                .eq("deleted", false)
        );

        return new PaginationWrapper<>(
                (int) result.getCurrent(),
                (int) result.getSize(),
                (int) result.getTotal(),
                result.getRecords()
        );
    }

    public StudentInfo getById(String id) {
        var result = studentInfoMapper.selectOne(new QueryWrapper<StudentInfo>()
                .select("id, student_number, name, major, grade, class, create_time, edit_time")
                .eq("id", id)
                .eq("deleted", false)
        );

        if (result == null) {
            throw new UserNotExistsException();
        }

        return result;
    }

    public List<StudentInfo> getByIds(List<String> idList) {
        if (idList.isEmpty()) {
            return new ArrayList<>();
        }

        var result = studentInfoMapper.selectList(new QueryWrapper<StudentInfo>()
                .select("id, student_number, name, major, grade, class")
                .in("id", idList)
                .eq("deleted", false)
        );

        if (result.isEmpty()) {
            throw new UserNotExistsException();
        }

        return result;
    }

    public StudentUserInfoDTO getByStudentNumber(String studentNumber) {
        var result = studentInfoMapper.selectOne(new QueryWrapper<StudentInfo>()
                .select("id, student_number, name, major, grade, class, create_time, edit_time")
                .eq("student_number", studentNumber)
                .eq("deleted", false)
        );

        if (result == null) {
            throw new UserNotExistsException();
        }

        var userBasic = userBasicService.getUserBasicById(result.getId());
        if (userBasic == null) {
            throw new UserNotExistsException();
        }

        return StudentUserInfoDTO.builder()
                .id(result.getId())
                .username(userBasic.getUsername())
                .number(result.getStudentNumber())
                .name(result.getName())
                .major(result.getMajor())
                .grade(result.getGrade())
                .clazz(result.getClazz())
                .build();
    }

    public StudentUserInfoDTO getInfoById(String id) {
        var result = studentInfoMapper.selectOne(new QueryWrapper<StudentInfo>()
                .select("id, student_number, name, major, grade, class, create_time, edit_time")
                .eq("id", id)
                .eq("deleted", false)
        );

        var userBasic = userBasicService.getUserBasicById(id);

        if (result == null || userBasic == null) {
            throw new UserNotExistsException();
        }

        return StudentUserInfoDTO.builder()
                .id(id)
                .username(userBasic.getUsername())
                .number(result.getStudentNumber())
                .name(result.getName())
                .major(result.getMajor())
                .grade(result.getGrade())
                .clazz(result.getClazz())
                .build();
    }

    @Transactional
    public StudentInfo add(BasicStudentInfoDto dto) {
        String id = Long.toString(YitIdHelper.nextId());

        StudentInfo studentInfo = new StudentInfo();
        studentInfo.setId(id);
        studentInfo.setStudentNumber(dto.studentNumber());
        studentInfo.setName(dto.studentName());
        studentInfo.setMajor(dto.major());
        studentInfo.setGrade(Integer.parseInt(dto.studentNumber().substring(0, 4)));
        studentInfo.setClazz("");
        studentInfo.setCreateTime(LocalDateTime.now());
        studentInfo.setEditTime(LocalDateTime.now());
        studentInfo.setDeleted(false);

        studentInfoMapper.insert(studentInfo);

        String username = dto.studentNumber();
        String password = "";
        userBasicService.addUser(id, username, password, (short) UserType.STUDENT.code);

        return studentInfo;
    }

    @Transactional
    public StudentInfo addWithAutoMajor(BasicStudentInfoDto dto) {
        var userBasic = userBasicService.getUserBasicByUsername(dto.studentNumber());
        if (userBasic != null) {
            return userInfoService.getAsStudent(userBasic.getId());
        }

        String id = Long.toString(YitIdHelper.nextId());

        StudentInfo studentInfo = new StudentInfo();
        studentInfo.setId(id);
        studentInfo.setStudentNumber(dto.studentNumber());
        studentInfo.setName(dto.studentName());

        var major = majorService.addMajorIfNotExists(dto.major());
        studentInfo.setMajor(major.getId());

        studentInfo.setGrade(Integer.parseInt(dto.studentNumber().substring(0, 4)));
        studentInfo.setClazz("unknown");
        studentInfo.setCreateTime(LocalDateTime.now());
        studentInfo.setEditTime(LocalDateTime.now());
        studentInfo.setDeleted(false);

        studentInfoMapper.insert(studentInfo);

        String username = dto.studentNumber();
        String password = "";
        userBasicService.addUser(id, username, password, (short) UserType.STUDENT.code);

        return studentInfo;
    }

    @Transactional
    public void delete(String id) {
        StudentInfo studentInfo = new StudentInfo();
        studentInfo.setDeleted(true);

        studentInfoMapper.update(studentInfo, new QueryWrapper<StudentInfo>()
                .eq("id", id)
        );

        userBasicService.deleteUser(id);
        userInfoService.invalidateCache(id);
    }

    @Transactional
    public StudentInfo update(String id, StudentUserInfoDTO dto) {
        StudentInfo studentInfo = this.getById(id);
        studentInfo.setStudentNumber(dto.id());
        studentInfo.setName(dto.name());
        studentInfo.setMajor(dto.major());
        studentInfo.setGrade(dto.grade());
        studentInfo.setClazz(dto.clazz());
        studentInfo.setEditTime(LocalDateTime.now());

        studentInfoMapper.update(studentInfo, new QueryWrapper<StudentInfo>()
                .eq("id", id)
                .eq("deleted", false)
        );

        UserBasicInfoDTO userBasicInfoDTO = new UserBasicInfoDTO(id, dto.username(), dto.password(), (short) UserType.STUDENT.code);
        userBasicService.updateUser(id, userBasicInfoDTO);

        return userInfoService.getAsStudent(id, false);
    }

    @Transactional
    public NewStudentDTO tryAddStudentFromOfficial(String studentNumber, String password) {
        var originInfo = officialUndergradStudentInfoService.getNewStudentInfo(studentNumber, password);

        String name = originInfo.getName();
        String majorName = originInfo.getMajor();
        String clazz = originInfo.getClazz();

        var studentInfo = studentInfoMapper.selectOne(new QueryWrapper<StudentInfo>()
                .eq("student_number", studentNumber)
                .eq("deleted", false)
        );

        String id;
        if (studentInfo == null) {
            id = Long.toString(YitIdHelper.nextId());
            studentInfo = StudentInfo.builder()
                    .id(id)
                    .studentNumber(studentNumber)
                    .name(name)
                    .grade(Integer.parseInt(studentNumber.substring(0, 4)))
                    .clazz(clazz)
                    .build();
        } else {
            id = studentInfo.getId();
        }

        var major = majorService.addMajorIfNotExists(majorName);
        studentInfo.setMajor(major.getId());
        studentInfoMapper.insertOrUpdate(studentInfo);

        var newUserBasic = userBasicService.addUser(id, studentNumber, password, (short) UserType.STUDENT.code);

        var newStudentInfo = userInfoService.getAsStudent(id, false);

        return new NewStudentDTO(newUserBasic, newStudentInfo);
    }
}
