package net.lensfrex.rd.service.user;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.github.yitter.idgen.YitIdHelper;
import net.lensfrex.rd.dao.mapper.user.TeacherInfoMapper;
import net.lensfrex.rd.data.dto.user.teacher.TeacherUserDTO;
import net.lensfrex.rd.data.dto.user.teacher.TeacherUserInfoDTO;
import net.lensfrex.rd.data.dto.user.UserBasicInfoDTO;
import net.lensfrex.rd.data.entity.table.TeacherInfo;
import net.lensfrex.rd.data.enums.UserType;
import net.lensfrex.rd.data.model.response.PaginationWrapper;
import net.lensfrex.rd.exception.user.UserNotExistsException;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;

@Service
public class TeacherUserService {
    private final TeacherInfoMapper teacherInfoMapper;

    private final UserBasicService userBasicService;
    private final UserInfoService userInfoService;

    public TeacherUserService(TeacherInfoMapper teacherInfoMapper,
                              UserBasicService userBasicService,
                              UserInfoService userInfoService) {
        this.teacherInfoMapper = teacherInfoMapper;
        this.userBasicService = userBasicService;
        this.userInfoService = userInfoService;
    }

    public PaginationWrapper<List<TeacherInfo>> list(int page, int pageSize) {
        Page<TeacherInfo> queryPage = new Page<>(page, pageSize);
        var result = teacherInfoMapper.selectPage(queryPage, new QueryWrapper<TeacherInfo>()
                .select("id, name, create_time, edit_time")
                .eq("deleted", false)
        );

        return new PaginationWrapper<>(
                (int) result.getCurrent(),
                (int) result.getSize(),
                (int) result.getTotal(),
                result.getRecords()
        );
    }

    public TeacherInfo getInfo(String id) {
        var result = teacherInfoMapper.selectOne(new QueryWrapper<TeacherInfo>()
                .select("id, name, create_time, edit_time")
                .eq("id", id)
                .eq("deleted", false)
        );

        if (result == null) {
            throw new UserNotExistsException();
        }

        return result;
    }

    public TeacherUserInfoDTO getUserInfoById(String id) {
        var result = teacherInfoMapper.selectOne(new QueryWrapper<TeacherInfo>()
                .select("id, name, create_time, edit_time")
                .eq("id", id)
                .eq("deleted", false)
        );

        var userBasic = userBasicService.getUserBasicById(id);

        if (result == null || userBasic == null) {
            throw new UserNotExistsException();
        }

        return new TeacherUserInfoDTO(id, userBasic.getUsername(), result.getName());
    }

    public List<TeacherInfo> getInfoById(List<String> ids) {
        if (ids.isEmpty()) {
            return new ArrayList<>();
        }

        return teacherInfoMapper.selectList(new QueryWrapper<TeacherInfo>()
                .select("id", "name")
                .in("id", ids)
                .eq("deleted", false)
        );
    }

    @Transactional
    public TeacherInfo add(TeacherUserDTO dto) {
        String id = Long.toString(YitIdHelper.nextId());

        TeacherInfo teacherInfo = new TeacherInfo();
        teacherInfo.setId(id);
        teacherInfo.setName(dto.teacherName());

        teacherInfoMapper.insert(teacherInfo);

        String username = (dto.username() == null) ? teacherInfo.getId() : dto.username();
        String password = dto.password();
        userBasicService.addUser(id, username, password, (short) UserType.TEACHER.code);

        return userInfoService.getAsTeacher(id, false);
    }

    @Transactional
    public void delete(String id) {
        TeacherInfo teacherInfo = new TeacherInfo();
        teacherInfo.setDeleted(true);

        teacherInfoMapper.update(teacherInfo, new QueryWrapper<TeacherInfo>()
                .eq("id", id)
        );

        userBasicService.deleteUser(id);
        userInfoService.invalidateCache(id);
    }

    @Transactional
    public TeacherInfo update(String id, TeacherUserDTO dto) {
        TeacherInfo teacherInfo = this.getInfo(id);
        teacherInfo.setName(dto.teacherName());
        teacherInfo.setEditTime(LocalDateTime.now());

        teacherInfoMapper.update(teacherInfo, new QueryWrapper<TeacherInfo>()
                .eq("id", id)
                .eq("deleted", false)
        );

        UserBasicInfoDTO userBasicInfoDTO = new UserBasicInfoDTO(id, dto.username(), dto.password(), (short) UserType.TEACHER.code);
        userBasicService.updateUser(id, userBasicInfoDTO);

        return userInfoService.getAsTeacher(id, false);
    }
}
